cd(@__DIR__)        # Move to this directory
println("Loading Modules")
using Ipopt         # Solver
using NLOptControl  # Problem definition and discretization
using PrettyPlots   # Plots for optimization problem
using Plots         # Custom plots
include("utils.jl") # Utilities

# =========== PROBLEM DEFINITION ==============================
println("Setup problem")
l    = 0.345
lact = 0.4
g    = 9.81
# Upper and lower constraints on control (daᵣ(t))
CL   = [...]    # FILL CONTROL (ȧᵣ(t)) LOWER CONSTRAINTS HERE
CU   = [...]    # FILL CONTROL (ȧᵣ(t)) UPPER CONSTRAINTS HERE
# Upper and lower constraints on states (p(t), dp(t), θ(t), dθ(t)), aᵣ(t))
XL   = [..., -Inf, -Inf, -Inf, ...]    # FILL STATE LOWER CONSTRAINTS HERE
XU   = [...,  Inf,  Inf,  Inf, ...]    # FILL STATE UPPER CONSTRAINTS HERE
# Initial and final constraints on states
X0   = [..., ..., ..., ..., 0]
XF   = [0, 0, 0, 0, NaN] # NaN means unspecified

# Create problem =======================================================================
prob  = define(numStates = 5, numControls=1, X0=X0, XF=XF, CL=CL, CU=CU, XL=XL, XU=XU)
# Give name to states and controls
states!(prob, [:p, :pd, :theta, :thetad, :ar], descriptions=["p(t)", "ṗ(t)", "θ(t)", "̇θ(t))", "aᵣ(t)"])
controls!(prob, [:ard], descriptions=["ȧᵣ(t)"])

# Setup model equations as julia **expressions** =======================================
dx    = Array{Expr}(5)
dx[1] = :( ..... ) # FILL IN EQUATIONS HERE
dx[2] = :( ..... ) # FILL IN EQUATIONS HERE
dx[3] = :( ..... ) # FILL IN EQUATIONS HERE
dx[4] = :( -$g/$l*sin(theta[j]) - 0.04*thetad[j] + 1/$l*cos(theta[j])*ar[j] ) # $ means insert value into expression
dx[5] = :( ard[j] )

## Add dynamics to model =============================================================
dynamics!(prob, dx)

# Obstracle constraints ================================================================
x_p = :( p[j]-$lact*sin(theta[j]) )
y_p = :( -$lact*cos(theta[j]) )
constraint_expr = :( (($x_p+0.3)/0.05)^2 + (($y_p+0.4)/0.3)^2 >= 1 )
constraints!(prob, [constraint_expr])

# =========== SOLVER AND OBJECTIVE DEFINITION ==========================================

N = 550 # Number of grid points in discretization
# Final time free, backward Euler discretization
configure!(prob; finalTimeDV=true, integrationScheme=:bkwEuler, N=N, solverSettings=(:print_level => 3))
# Set objective to minimize final time
@NLobjective(prob.ocp.mdl, Min, prob.ocp.tf)

println("Setting initial guess")
# Initial guess =========================================================================
initialdata, header = read_initial_guess(newinit=true) # juliainit=true means guess generated by julia
t_guess, states_guess, controls_guess, tf_guess = initialdata[:,1], initialdata[:,2:6], initialdata[:,7:7], initialdata[end,1]
set_guess!(prob, t_guess, states_guess, controls_guess, tf_guess)

# Solve ================================================================================
println("Optimizing")
@time optimize!(prob)
println("Done")
println("Final time: ", getvalue(prob.ocp.tf))

# Write solution to file
println("Saving solution to \"pendulum_result.csv\"")
dfs = Array(prob.r.ocp.dfs[1]) # All states
writedlm("pendulum_result.csv", ["time" "p" "p_dot" "theta" "theta_dot" "a_ref" "a_ref_dot"; dfs], " ")


### ============================= PLOTTING AND EXTRAS =================================

# Plot all states and controls with bounds ============================================
function plotAll(prob)
    println("Plotting all states")
    allPlots(prob)
end

# Plot constraint circle ==============================================================
function plot_trajectory()
    println("Plotting constrint and end of pendulum")
    global dfs, lact
    p       = dfs[:,2]
    theta   = dfs[:,4]

    plt = plot([-1,1], [0,0], ylims=(-0.8,0.8), xlims=(-1.3,0.1), aspect_ratio=1, lab="")
    plot!(-0.3-0.05*sin.(0:0.1:2π), -0.4-0.3*cos.(0:0.1:2pi), lab="")
    scatter!([p.-lact*sin.(theta)], [-lact*cos.(theta)], m=(:circle, 3.0), lab="")
    return plt
end

# Crate gif
function create_animation()
    global N, dfs, lact, prob
    p       = dfs[:,2]
    theta   = dfs[:,4]
    # Create Animation ====================================================================
    println("Creating animation")
    anim = @animate for k = (-20):6:(N+20)
        i = k > N ? N : (k < 1 ? 1 : k)                                                 # Pause before and after
        plot([-1,1], [0,0], ylims=(-0.8,0.8), xlims=(-1.3,0.1), aspect_ratio=1, lab="") # Beam
        plot!([p[i], p[i]-lact*sin(theta[i])], [0.0, -lact*cos(theta[i])], lab="")      # Pendulum
        scatter!([p[i],], [0.0], m=:star, lab="")                                       # Pendulum base
        plot!(-0.3-0.05*sin.(0:0.1:2pi), -0.4-0.3*cos.(0:0.1:2pi), lab="")              # Constraint
        scatter!([p[1:i]-lact*sin.(theta[1:i])], [-lact*cos.(theta[1:i])], m=(:circle,3.0), lab="") # Mass
    end
    # 30 Fps at N = 550
    gif(anim, "/tmp/pendulum.gif", fps = round(Int, length((-20):6:(N+20))/getvalue(prob.ocp.tf)))
    println("Animation saved as \"/tmp/pendulum.gif\"")
end

println("All done")
println("Run \"plotAll(prob)\" to plot all states")
println("Run \"plot_trajectory()\" to see trajectory and obstacle")
println("Run \"create_animation()\" to generate an animation")
